using System;
using System.Text.RegularExpressions;
using System.Runtime.InteropServices;
using Tao.Sdl;

using System.Reflection;
[assembly: AssemblyTitle("Swarm.NET")]
[assembly: AssemblyVersion("1.1")]
[assembly: AssemblyCopyright("(C) 2007 Chip Black")]

// WHERE IS YOUR OO GOD NOW?
public struct global {
	public static Random random;
}

public enum ClearOpt { MINRECT, FULLCLEAR }

class Swarm {
	IntPtr canvas;
	int mx, my;
	int width, height;
	int fill;
	Pixel[] pixels;
	IntPtr canvasformat;
	ClearOpt clearopt;

	class Pixel {
		public double x, y;
		public double vx, vy;
		int lastupdate;
		int stroke;

		public Pixel(Swarm parent) {
			x = global.random.NextDouble() * parent.width;
			y = global.random.NextDouble() * parent.height;
			vx = global.random.NextDouble() * 10.0 - 5.0;
			vy = global.random.NextDouble() * 10.0 - 5.0;
			lastupdate = Sdl.SDL_GetTicks();
			stroke = 0x000000FF;
		}

		public void accel(double x, double y) {
			vx += x;
			if (vx > 100.0)
				vx = 100.0;
			else if (vx < -100.0)
				vx = -100.0;
			vy += y;
			if (vy > 100.0)
				vy = 100.0;
			else if (vy < -100.0)
				vy = -100.0;
		}

		public void update() {
			int now = Sdl.SDL_GetTicks();
			double dt = (now - lastupdate) / 75.0;
			x += vx * dt;
			y += vy * dt;
			lastupdate = now;
		}

		public void draw(IntPtr context) {
			SdlGfx.aalineColor(context, 
					(short) x, (short) y,
					(short) (x - vx / 4.0),
					(short) (y - vy / 4.0),
					stroke);
		}
	}

	public Swarm(short w, short h, int npixels) {
		mx = w / 2;
		my = h / 2;
		width = w;
		height = h;

		resize(w,h);

		// Initialize pixel objects
		pixels = new Pixel[npixels];
		for (int i = 0; i < npixels; i++) {
			pixels[i] = new Pixel(this);
		}
	}

	public void mousemove(int x, int y) {
		mx = x;
		my = y;
	}

	public void mousedown() {
		// Apply force to each pixel outwards from the mouse
		foreach (Pixel p in pixels) {
			// Get the direction
			double dx = p.x - mx;
			double dy = p.y - my;
			double mag = Math.Sqrt(dx*dx + dy*dy);
			dx /= mag;
			dy /= mag;

			// Apply force.
			p.accel(dx * 25.0, dy * 25.0);
		}
	}

	// VROOM!
	public void engine() {
		// Update all pixel objects
		foreach (Pixel p in pixels) {
			// Get distance
			double dx = mx - p.x;	// MXPX isn't realy all that great
			double dy = my - p.y;
			double d = Math.Sqrt(dx*dx + dy*dy);

			// Apply acceleration to pixels towards the mouse cursor
			double ax = dx / 50.0;
			double ay = dy / 50.0;

			// Apply a drag proportional to the distance from the cursor
			if (d > 25.0) {
				ax += -p.vx * 0.1 * (d - 25.0) / 100.0;
				ay += -p.vy * 0.1 * (d - 25.0) / 100.0;
			}

			// And a "chaotic" acceleration inversely proportional
			// to the distance from the cursor
			ax += (global.random.NextDouble() * 40.0 - 20.0) / d;
			ay += (global.random.NextDouble() * 40.0 - 20.0) / d;

			p.accel(ax,ay);
			p.update();
			p.draw(canvas);
		}

		SdlGfx.rectangleColor(canvas,
				(short) (mx-5), (short) (my-5),
				(short) (mx+5), (short) (my+5),
				0x000000FF);

		Sdl.SDL_Flip(canvas);

		// Clean up after ourselves (this is easier to do here than in the next
		// iteration since we'd have to save state)
		Sdl.SDL_Rect r;
		switch(clearopt) {
		case ClearOpt.MINRECT:
			int x1 = mx - 5, y1 = my - 5, x2 = mx + 5, y2 = my + 5;
			foreach (Pixel p in pixels) {
				int x = (int) p.x;
				int y = (int) p.y;
				int vx = (int) Math.Ceiling(Math.Abs(p.vx / 4.0));
				int vy = (int) Math.Ceiling(Math.Abs(p.vy / 4.0));
				if (x - vx < x1)
					x1 = x - vx;
				else if (x + vx > x2)
					x2 = x + vx;
				if (y - vy < y1)
					y1 = y - vy;
				else if (y + vy > y2)
					y2 = y + vy;
			}
			x2++; y2++;
			r = new Sdl.SDL_Rect((short) x1, (short) y1,
								 (short) (x2-x1), (short) (y2-y1));
			Sdl.SDL_FillRect(canvas, ref r, fill);
			//SdlGfx.rectangleRGBA(canvas, (short)x1, (short)y1, (short)x2, (short)y2, 0,0,0,255);
			break;
		case ClearOpt.FULLCLEAR:
			r = new Sdl.SDL_Rect(0,0,(short)width,(short)height);
			Sdl.SDL_FillRect(canvas, ref r, fill);
			break;
		}
	}

	public void resize(short w, short h) {
		width = w;
		height = h;
		canvas = Sdl.SDL_SetVideoMode(w, h, 0,
				Sdl.SDL_HWSURFACE | Sdl.SDL_DOUBLEBUF |
				Sdl.SDL_RESIZABLE);
		if (canvas.ToInt32() == 0) {
			Console.WriteLine("Could not initialize SDL Video. Hurgh!");
			Environment.Exit(1);
		}
		unsafe {
			Sdl.SDL_Surface *s = (Sdl.SDL_Surface*) canvas.ToPointer();
			canvasformat = s->format;
		}
		Sdl.SDL_Rect r = new Sdl.SDL_Rect(0,0,w,h);
		fill = Sdl.SDL_MapRGB(canvasformat, 128,128,128);
		Sdl.SDL_FillRect(canvas, ref r, fill);
	}

	public static void Main(string[] Args) {
		int fpstime;
		int fpsupdate = 0;
		int npixels = 100;
		int timedemo = 0;
		ClearOpt clearopt = ClearOpt.MINRECT;

		global.random = new Random();
		Sdl.SDL_Init(Sdl.SDL_INIT_VIDEO);

		Regex option = new Regex(@"^-([A-Za-z]+)$");
		for (int i = 0; i < Args.Length; i++) {
			Match m = option.Match(Args[i]);
			if (m.Success) {
				switch (m.Groups[1].Value) {
				case "timedemo":
					try {
						timedemo = System.Convert.ToInt32(Args[++i]);
					} catch (FormatException) {
						Console.WriteLine("Invalid argument to -timedemo.");
					}
					break;
				case "fullclear":
					clearopt = ClearOpt.FULLCLEAR;
					break;
				}
			} else {
				try {
					npixels = System.Convert.ToInt32(Args[i]);
				} catch (FormatException) {
					Console.WriteLine("Invalid argument: " + Args[i]);
				}
			}
		}

		if (timedemo > 0) {
			Console.WriteLine("Timedemo enabled for " + timedemo.ToString() + " frames.");
		}

		Swarm swarm = new Swarm(640, 480, npixels);
		swarm.clearopt = clearopt;
		Sdl.SDL_WM_SetCaption("Swarm","");

		// Initialize framerate manager
		SdlGfx.FPSmanager fpsobj;
		IntPtr fpsman;
		unsafe { fpsman = new IntPtr((void*) &fpsobj); }
		SdlGfx.SDL_initFramerate(fpsman);
		SdlGfx.SDL_setFramerate(fpsman, 60);

		fpstime = Sdl.SDL_GetTicks();
		bool running = true;
		while (running) {
			Sdl.SDL_Event e;
			while (Sdl.SDL_PollEvent(out e) == 1) {
				switch(e.type) {
				case Sdl.SDL_MOUSEBUTTONDOWN:
					swarm.mousedown();
					break;
				case Sdl.SDL_MOUSEMOTION:
					swarm.mousemove(e.motion.x, e.motion.y);
					break;
				case Sdl.SDL_KEYDOWN:
					if (e.key.keysym.sym == Sdl.SDLK_q ||
					    e.key.keysym.sym == Sdl.SDLK_ESCAPE) {
						goto case Sdl.SDL_QUIT;
					}
					break;
				case Sdl.SDL_VIDEORESIZE:
					swarm.resize((short)e.resize.w, (short)e.resize.h);
					fpstime = Sdl.SDL_GetTicks();
					fpsupdate = 0;
					break;
				case Sdl.SDL_QUIT:
					running = false;
					return;
				}
			}
			swarm.engine();
			if (timedemo > 0) {
				if (++fpsupdate == timedemo) {
					double delta = (Sdl.SDL_GetTicks() - fpstime) / 1000.0;
					Console.WriteLine("{0} frames in {1} seconds = {2} FPS", timedemo, delta, timedemo / delta);
					running = false;
				}
			} else {
				if (++fpsupdate == 100) {
					int now = Sdl.SDL_GetTicks();
					double fps = 100000.0 / (now - fpstime);
					Sdl.SDL_WM_SetCaption("Swarm [" + fps.ToString("F1")
						+ " FPS, " + npixels.ToString() + " pixels]","");
					fpsupdate = 0;
					fpstime = now;
				}
			}
			if (timedemo == 0)
				SdlGfx.SDL_framerateDelay(fpsman);
		}
	}
}

// vim:set ts=4 sw=4 tw=0:
